# Please note, this is just for compability

from snudda.detect.prune import SnuddaPrune

print("SnuddaPrune is now available when doing import snudda.detect")
print("Please update to `from snudda.detect import SnuddaPrune`")
